import re
from threading import Thread

from slackclient import SlackClient

from utils import _redirect_output

_handlers = {}

class Yasuf:
    def __init__(self, token, channel=None):
        self.sc = SlackClient(token)
        self.start_time = None
        self.default_channel = channel

    def run(self):
        """ Creates a thread that monitors Slack messages and returns """

        Thread(target=self._run)

    def _run(self):
        if self.sc.rtm_connect():
            self._synchronize_time()
            for message in self._get_message():
                if message.get('type') == 'message':
                    for trigger, fun in _handlers.items():
                        print(message)
                        match = trigger.match(message.get('text'))
                        if match:
                            stdout, ret_val = fun.execute(match.groups())
                            output = '\n'.join(stdout.readlines()) + str(ret_val)
                            self.sc.api_call('chat.postMessage',
                                             channel=self.default_channel,
                                             text=output)

    def _get_message(self):
        """ Yields new messages """

        while True:
            message = self.sc.rtm_read()
            if message and float(message[0].get('ts', 0)) > self.start_time:
                yield message[0]

    def _synchronize_time(self):
        """ Sends a message and records channel local time.
            rtm_read seems to resend old messages when first connecting. We want
            to avoid reacting to those. """

        self.sc.api_call('chat.postMessage', channel='#general', text='Hello!')
        for message in self._get_message():
            print(message)
            if message.get('subtype') == 'bot_message' and message['text'] == 'Hello!':
                self.start_time = float(message['ts'])
                print('Synchronized at {}'.format(self.start_time))

    class handle():
        def __init__(self, trigger, channel=None, types=None):
            self.regexp = re.compile(trigger)
            self.types = types

        def __call__(self, f):
            self.fun = _redirect_output(f)
            _handlers[self.regexp] = self

        def execute(self, groups):
            if self.types is not None:
                assert len(self.types) == len(groups)
                params = [t(g) for t, g in zip(self.types, groups)]
            else:
                params = groups
            return self.fun(*params)
